#include "Util-pkg.h"
#include "P3d-pkg.h"
#include "Localpath-pkg.h"
#include "Graphic-pkg.h"
#ifdef LIGHT_PLANNER
#include "ManipulationArmData.hpp"
#endif

#define DEBUG_LEVEL 0

typedef enum {
    A1,
    A2,
    B,
    C1,
    C2,
    D,
    E1,
    E2,
    G,
    H1,
    H2,
    F
} times_index;

typedef enum {
    A,
    C,
    E,
    H
} signes_index;

double sign(double d){
    if(d<0) return -1;
    if(d>0) return  1;
    return 0;
}

double get_v_C3_tC1(p3d_rob *robotPt,
                    p3d_quadspline_data *quadspline_specificPt, int k,
                    double a_A3_tA1, double tB){

    lm_quadspline_str *quadspline_paramPt;
    double v0,a0,jmax,jmax_2,smax;
    double jmaxA,diffA,sA,tA1,tA1_2,tA1_3,tA2,tA2_2;
    double jmaxC,diffC,sC,tC1,tC1_2,tC1_3,tC2,tC2_2;
    double v_C3_tC1;

    quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in get_v_C3_tC1 : quadspline_paramPt is NULL\n");
        return 0;
    }

    if(k<3){
        v0=quadspline_specificPt->q_init[quadspline_paramPt->vx_dof+k];
        a0=quadspline_specificPt->q_init[quadspline_paramPt->ax_dof+k];
    }
    else{
        v0=0;
        a0=0;
    }
    jmax=quadspline_paramPt->jmax;
    jmax_2=jmax*jmax;
    smax=quadspline_paramPt->smax;

    sA=sign(a_A3_tA1-a0);
    jmaxA=-sA*smax*(a0-a_A3_tA1);
    if(jmax_2<=jmaxA){
        jmaxA=jmax;
    }
    else{
        jmaxA=sqrt(jmaxA);
    }

    tA1=jmaxA/smax;
    tA1_2=tA1*tA1;
    tA1_3=tA1_2*tA1;

    diffA=-sA*(a0-a_A3_tA1)-jmaxA*jmaxA/smax;
    if(diffA>0){
        tA2=-(sA*smax*tA1_2+a0-a_A3_tA1)/(sA*smax*tA1);
    }
    else{
        tA2=0;
    }

    tA2_2=tA2*tA2;

    sC=-sign(a_A3_tA1);
    jmaxC=-smax*(sA*smax*tA1_2+sA*smax*tA2*tA1+a0)*sC;
    if(jmax_2<=jmaxC){
        jmaxC=jmax;
    }
    else{
        jmaxC=sqrt(jmaxC);
    }

    tC1=jmaxC/smax;
    tC1_2=tC1*tC1;
    tC1_3=tC1_2*tC1;

    diffC=(sC*jmaxC*jmaxC)/smax+a0+sA*smax*tA1*(tA1+tA2);
    if(sC*diffC<0){
        tC2=-sC*diffC/jmaxC;
    }
    else{
        tC2=0;
    }

    tC2_2=tC2*tC2;

    quadspline_specificPt->signes[A][k]=sA;
    quadspline_specificPt->times[A1][k]=tA1;
    quadspline_specificPt->times[A2][k]=tA2;

    quadspline_specificPt->signes[C][k]=sC;
    quadspline_specificPt->times[C1][k]=tC1;
    quadspline_specificPt->times[C2][k]=tC2;

    v_C3_tC1=sA*smax*tA1_3+(3*sA*smax*tA1_2*tA2)/2+2*sA*smax*tA1_2*tC1
             +sA*smax*tA1_2*tC2+sA*smax*tB*tA1_2+(sA*smax*tA1*tA2_2)/2
             +2*sA*smax*tA1*tA2*tC1+sA*smax*tA1*tA2*tC2+sA*smax*tB*tA1*tA2
             +2*a0*tA1+a0*tA2+sC*smax*tC1_3+(3*sC*smax*tC1_2*tC2)/2
             +(sC*smax*tC1*tC2_2)/2+2*a0*tC1+a0*tC2+v0+a0*tB;

    return v_C3_tC1;
}

double get_v_E1_0(p3d_rob *robotPt, p3d_quadspline_data *quadspline_specificPt,
                  int k, double a_H1_0, double tG){

    lm_quadspline_str *quadspline_paramPt;
    double vF,aF,jmax,jmax_2,smax;
    double jmaxE,diffE,sE,tE1,tE1_2,tE1_3,tE2,tE2_2;
    double jmaxH,diffH,sH,tH1,tH1_2,tH1_3,tH2,tH2_2;
    double v_E1_0;

    quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in get_v_E1_0 : quadspline_paramPt is NULL\n");
        return 0;
    }

    if(k<3){
        vF=quadspline_specificPt->q_end[quadspline_paramPt->vx_dof+k];
        aF=quadspline_specificPt->q_end[quadspline_paramPt->ax_dof+k];
    }
    else{
        vF=0;
        aF=0;
    }
    jmax=quadspline_paramPt->jmax;
    jmax_2=jmax*jmax;
    smax=quadspline_paramPt->smax;

    sH=sign(aF-a_H1_0);
    jmaxH=sH*smax*(aF-a_H1_0);
    tH2=0;
    if(jmax_2<=jmaxH){
        jmaxH=jmax;
        tH2=-1;
    }
    else{
        jmaxH=sqrt(jmaxH);
    }

    tH1=jmaxH/smax;
    tH1_2=tH1*tH1;
    tH1_3=tH1_2*tH1;

    if(tH2<0){
        diffH=aF-a_H1_0-jmaxH*jmaxH*sH/smax;
        tH2=sH*diffH/jmaxH;
    }

    tH2_2=tH2*tH2;

    sE=-sign(a_H1_0);
    jmaxE=fabs(a_H1_0)*smax;
    if(jmax_2<=jmaxE){
        jmaxE=jmax;
        tE2=-1;
    }
    else{
        jmaxE=sqrt(jmaxE);
        tE2=0;
    }

    tE1=jmaxE/smax;
    tE1_2=tE1*tE1;
    tE1_3=tE1_2*tE1;

    if(tE2<0){
        diffE=sE*smax*tE1_2+a_H1_0;
        tE2=-sE*diffE/jmaxE;
    }

    tE2_2=tE2*tE2;

    quadspline_specificPt->signes[E][k]=sE;
    quadspline_specificPt->times[E1][k]=tE1;
    quadspline_specificPt->times[E2][k]=tE2;

    quadspline_specificPt->signes[H][k]=sH;
    quadspline_specificPt->times[H1][k]=tH1;
    quadspline_specificPt->times[H2][k]=tH2;

    v_E1_0=vF-2*aF*tE1-aF*tE2-aF*tG-2*aF*tH1-aF*tH2-sE*smax*tE1_3
         +sH*smax*tH1_3-(sE*smax*tE1*tE2_2)/2-(3*sE*smax*tE1_2*tE2)/2
         +2*sH*smax*tE1*tH1_2+sH*smax*tE2*tH1_2+sH*smax*tG*tH1_2
         +(sH*smax*tH1*tH2_2)/2+(3*sH*smax*tH1_2*tH2)/2
         +2*sH*smax*tE1*tH1*tH2+sH*smax*tE2*tH1*tH2+sH*smax*tG*tH1*tH2;

    return v_E1_0;

}

double get_a_A3_tA1_tB(p3d_rob *robotPt,
                       p3d_quadspline_data *quadspline_specificPt, int k,
                       double v_C3_tC1){

    lm_quadspline_str *quadspline_paramPt;
    double v0,a0,amax,jmax,smax,tB,vMin,vMid,vMax,aMin,aMax,s,dB,v,tBMin,tBMax;
    double a_A3_tA1;

    quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in get_a_A3_tA1_tB : quadspline_paramPt is NULL\n");
        return 0;
    }

    if(k<3){
        v0=quadspline_specificPt->q_init[quadspline_paramPt->vx_dof+k];
        a0=quadspline_specificPt->q_init[quadspline_paramPt->ax_dof+k];
    }
    else{
        v0=0;
        a0=0;
    }
    amax=quadspline_paramPt->amax;
    jmax=quadspline_paramPt->jmax;
    smax=quadspline_paramPt->smax;

    tB=0;
    vMin=get_v_C3_tC1(robotPt,quadspline_specificPt,k,-amax,0);
    vMax=get_v_C3_tC1(robotPt,quadspline_specificPt,k,amax,0);
    if(vMin<=v_C3_tC1 && v_C3_tC1<=vMax){
        aMin=-amax;
        aMax=amax;
        do{
            a_A3_tA1=(aMax+aMin)/2;
            vMid=get_v_C3_tC1(robotPt,quadspline_specificPt,k,a_A3_tA1,tB);
            if(vMid<v_C3_tC1){
                aMin=a_A3_tA1;
            }
            else{
                aMax=a_A3_tA1;
            }
        }while(fabs(vMid-v_C3_tC1)>1e-10);
        quadspline_specificPt->times[B][k]=tB;
        return a_A3_tA1;
    }
    else{
        if(v_C3_tC1<vMin){
            a_A3_tA1=-amax;
            s=-1;
        }
        else{
            a_A3_tA1=amax;
            s=1;
        }
        dB=0.5;
        do{
            tB=tB+dB;
            v=get_v_C3_tC1(robotPt,quadspline_specificPt,k,a_A3_tA1,tB);
        }while(s*(v_C3_tC1-v)>0);
        tBMax=tB;
        tBMin=tB-dB;
        do{
            tB=(tBMin+tBMax)/2;
            v=get_v_C3_tC1(robotPt,quadspline_specificPt,k,a_A3_tA1,tB);
            if(s*(v_C3_tC1-v)>0){
                tBMin=tB;
            }
            else{
                tBMax=tB;
            }
        }while(fabs(v_C3_tC1-v)>1e-10);
        quadspline_specificPt->times[B][k]=tB;
        return a_A3_tA1;
    }
}

double get_a_H1_0_tG(p3d_rob *robotPt,
                     p3d_quadspline_data *quadspline_specificPt, int k,
                     double v_E1_0){

    lm_quadspline_str *quadspline_paramPt;
    double vF,aF,amax,jmax,smax,tG,vMin,vMid,vMax,aMin,aMax,s,dG,v,v1,v2,tGMin;
    double a_H1_0,tGMax;

    quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in get_a_H1_0_tG : quadspline_paramPt is NULL\n");
        return 0;
    }

    if(k<3){
        vF=quadspline_specificPt->q_end[quadspline_paramPt->vx_dof+k];
        aF=quadspline_specificPt->q_end[quadspline_paramPt->ax_dof+k];
    }
    else{
        vF=0;
        aF=0;
    }
    amax=quadspline_paramPt->amax;
    jmax=quadspline_paramPt->jmax;
    smax=quadspline_paramPt->smax;

    tG=0;
    v1=get_v_E1_0(robotPt,quadspline_specificPt,k,-amax,0);
    v2=get_v_E1_0(robotPt,quadspline_specificPt,k,amax,0);
    vMin=v1;
    if(v2<vMin) vMin=v2;
    vMax=v1;
    if(v2>vMax) vMax=v2;
    if(v1<v2){
        aMin=-amax;
        aMax=amax;
    }
    else{
        aMin=amax;
        aMax=-amax;
    }
    if(vMin<=v_E1_0 && v_E1_0<=vMax){
        do{
            a_H1_0=(aMax+aMin)/2;
            vMid=get_v_E1_0(robotPt,quadspline_specificPt,k,a_H1_0,tG);
            if(vMid<v_E1_0){
                aMin=a_H1_0;
            }
            else{
                aMax=a_H1_0;
            }
        }while(fabs(vMid-v_E1_0)>1e-10);
        quadspline_specificPt->times[G][k]=tG;
        return a_H1_0;
    }
    else{
        if(v_E1_0>vMax){
            a_H1_0=aMax;
            s=1;
        }
        else{
            a_H1_0=aMin;
            s=-1;
        }
        dG=0.5;
        do{
            tG=tG+dG;
            v=get_v_E1_0(robotPt,quadspline_specificPt,k,a_H1_0,tG);
        }while(s*(v_E1_0-v)>0);
        tGMax=tG;
        tGMin=tG-dG;
        do{
            tG=(tGMin+tGMax)/2;
            v=get_v_E1_0(robotPt,quadspline_specificPt,k,a_H1_0,tG);
            if(s*(v_E1_0-v)>0){
                tGMin=tG;
            }
            else{
                tGMax=tG;
            }
        }while(fabs(v_E1_0-v)>1e-10);
        quadspline_specificPt->times[G][k]=tG;
        return a_H1_0;
    }
}

void get_xC_xE(p3d_rob *robotPt, p3d_quadspline_data *quadspline_specificPt,
               int k, double vMid, double *xC, double *xE){

    lm_quadspline_str *quadspline_paramPt;
    double amax,jmax,smax,x0,xF,v0,vF,a0,aF;
    double tA1,tA2,sA,tB,tC1,tC2,sC,tD,tE1,tE2,sE,tG,tH1,tH2,sH;
    double tA1_2,tA1_3,tA1_4,tA2_2,tA2_3,tB_2,tC1_2,tC1_3,tC1_4,tC2_2,tC2_3;
    double tE1_2,tE1_3,tE1_4,tE2_2,tE2_3,tG_2,tH1_2,tH1_3,tH1_4,tH2_2,tH2_3;

    quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in get_xC_xE : quadspline_paramPt is NULL\n");
        return;
    }

    if(k<3){
        x0=quadspline_specificPt->q_init[quadspline_paramPt->x_dof+k];
        v0=quadspline_specificPt->q_init[quadspline_paramPt->vx_dof+k];
        a0=quadspline_specificPt->q_init[quadspline_paramPt->ax_dof+k];
        xF=quadspline_specificPt->q_end[quadspline_paramPt->x_dof+k];
        vF=quadspline_specificPt->q_end[quadspline_paramPt->vx_dof+k];
        aF=quadspline_specificPt->q_end[quadspline_paramPt->ax_dof+k];
    }
    else{
        x0=quadspline_specificPt->q_init[quadspline_paramPt->rz_dof];
        v0=0;
        a0=0;
        xF=quadspline_specificPt->q_end[quadspline_paramPt->rz_dof];
        vF=0;
        aF=0;
    }
    amax=quadspline_paramPt->amax;
    jmax=quadspline_paramPt->jmax;
    smax=quadspline_paramPt->smax;

    get_a_A3_tA1_tB(robotPt,quadspline_specificPt,k,vMid);
    get_a_H1_0_tG(robotPt,quadspline_specificPt,k,vMid);

    sA=quadspline_specificPt->signes[A][k];
    sC=quadspline_specificPt->signes[C][k];
    sE=quadspline_specificPt->signes[E][k];
    sH=quadspline_specificPt->signes[H][k];

    tA1=quadspline_specificPt->times[A1][k];
    tA2=quadspline_specificPt->times[A2][k];
    tB=quadspline_specificPt->times[B][k];
    tC1=quadspline_specificPt->times[C1][k];
    tC2=quadspline_specificPt->times[C2][k];
    tD=quadspline_specificPt->times[D][k];
    tE1=quadspline_specificPt->times[E1][k];
    tE2=quadspline_specificPt->times[E2][k];
    tG=quadspline_specificPt->times[G][k];
    tH1=quadspline_specificPt->times[H1][k];
    tH2=quadspline_specificPt->times[H2][k];

    tA1_2=tA1*tA1;
    tA1_3=tA1_2*tA1;
    tA1_4=tA1_3*tA1;

    tA2_2=tA2*tA2;
    tA2_3=tA2_2*tA2;

    tB_2=tB*tB;

    tC1_2=tC1*tC1;
    tC1_3=tC1_2*tC1;
    tC1_4=tC1_3*tC1;

    tC2_2=tC2*tC2;
    tC2_3=tC2_2*tC2;

    tE1_2=tE1*tE1;
    tE1_3=tE1_2*tE1;
    tE1_4=tE1_3*tE1;

    tE2_2=tE2*tE2;
    tE2_3=tE2_2*tE2;

    tG_2=tG*tG;

    tH1_2=tH1*tH1;
    tH1_3=tH1_2*tH1;
    tH1_4=tH1_3*tH1;

    tH2_2=tH2*tH2;
    tH2_3=tH2_2*tH2;

   *xC=(7*sA*smax*tA1_4)/12+(7*sA*smax*tA1_3*tA2)/6+sA*smax*tA1_3*tB
      +2*sA*smax*tA1_3*tC1+sA*smax*tA1_3*tC2+(3*sA*smax*tA1_2*tA2_2)/4
      +(3*sA*smax*tA1_2*tA2*tB)/2+3*sA*smax*tA1_2*tA2*tC1
      +(3*sA*smax*tA1_2*tA2*tC2)/2+(sA*smax*tA1_2*tB_2)/2
      +2*sA*smax*tA1_2*tB*tC1+sA*smax*tA1_2*tB*tC2+2*sA*smax*tA1_2*tC1_2
      +2*sA*smax*tA1_2*tC1*tC2+(sA*smax*tA1_2*tC2_2)/2+2*a0*tA1_2
      +(sA*smax*tA1*tA2_3)/6+(sA*smax*tA1*tA2_2*tB)/2
      +sA*smax*tA1*tA2_2*tC1+(sA*smax*tA1*tA2_2*tC2)/2
      +(sA*smax*tA1*tA2*tB_2)/2+2*sA*smax*tA1*tA2*tB*tC1
      +sA*smax*tA1*tA2*tB*tC2+2*sA*smax*tA1*tA2*tC1_2
      +2*sA*smax*tA1*tA2*tC1*tC2+(sA*smax*tA1*tA2*tC2_2)/2
      +2*a0*tA1*tA2+2*a0*tA1*tB+4*a0*tA1*tC1+2*a0*tA1*tC2+2*v0*tA1
      +(a0*tA2_2)/2+a0*tA2*tB+2*a0*tA2*tC1+a0*tA2*tC2+v0*tA2+(a0*tB_2)/2
      +2*a0*tB*tC1+a0*tB*tC2+v0*tB+(7*sC*smax*tC1_4)/12
      +(7*sC*smax*tC1_3*tC2)/6+(3*sC*smax*tC1_2*tC2_2)/4+2*a0*tC1_2
      +(sC*smax*tC1*tC2_3)/6+2*a0*tC1*tC2+2*v0*tC1+(a0*tC2_2)/2+v0*tC2+x0;

   *xE=(7*sE*smax*tE1_4)/12+(7*sE*smax*tE1_3*tE2)/6
      +(3*sE*smax*tE1_2*tE2_2)/4-2*sH*smax*tE1_2*tH1_2
      -2*sH*smax*tE1_2*tH1*tH2+2*aF*tE1_2+(sE*smax*tE1*tE2_3)/6
      -2*sH*smax*tE1*tE2*tH1_2-2*sH*smax*tE1*tE2*tH1*tH2+2*aF*tE1*tE2
      -2*sH*smax*tE1*tG*tH1_2-2*sH*smax*tE1*tG*tH1*tH2+2*aF*tE1*tG
      -2*sH*smax*tE1*tH1_3-3*sH*smax*tE1*tH1_2*tH2-sH*smax*tE1*tH1*tH2_2
      +4*aF*tE1*tH1+2*aF*tE1*tH2-2*vF*tE1-(sH*smax*tE2_2*tH1_2)/2
      -(sH*smax*tE2_2*tH1*tH2)/2+(aF*tE2_2)/2-sH*smax*tE2*tG*tH1_2
      -sH*smax*tE2*tG*tH1*tH2+aF*tE2*tG-sH*smax*tE2*tH1_3
      -(3*sH*smax*tE2*tH1_2*tH2)/2-(sH*smax*tE2*tH1*tH2_2)/2+2*aF*tE2*tH1
      +aF*tE2*tH2-vF*tE2-(sH*smax*tG_2*tH1_2)/2-(sH*smax*tG_2*tH1*tH2)/2
      +(aF*tG_2)/2-sH*smax*tG*tH1_3-(3*sH*smax*tG*tH1_2*tH2)/2
      -(sH*smax*tG*tH1*tH2_2)/2+2*aF*tG*tH1+aF*tG*tH2-vF*tG
      -(7*sH*smax*tH1_4)/12-(7*sH*smax*tH1_3*tH2)/6
      -(3*sH*smax*tH1_2*tH2_2)/4+2*aF*tH1_2-(sH*smax*tH1*tH2_3)/6
      +2*aF*tH1*tH2-2*vF*tH1+(aF*tH2_2)/2-vF*tH2+xF;

    quadspline_specificPt->times[F][k]=2*tA1+tA2+tB+2*tC1+tC2
                                      +2*tE1+tE2+tG+2*tH1+tH2;

}

double getVopt(p3d_rob *robotPt, p3d_quadspline_data *quadspline_specificPt,
               int k){

    lm_quadspline_str *quadspline_paramPt;
    double vmax,amax,jmax,smax,x0,xF,v0,vF,a0,aF;
    double dv,tD,vMax,xC,xE,vMaxMin,vMaxMax;

    quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in getVopt : quadspline_paramPt is NULL\n");
        return 0;
    }

    if(k<3){
        x0=quadspline_specificPt->q_init[quadspline_paramPt->x_dof+k];
        v0=quadspline_specificPt->q_init[quadspline_paramPt->vx_dof+k];
        a0=quadspline_specificPt->q_init[quadspline_paramPt->ax_dof+k];
        xF=quadspline_specificPt->q_end[quadspline_paramPt->x_dof+k];
        vF=quadspline_specificPt->q_end[quadspline_paramPt->vx_dof+k];
        aF=quadspline_specificPt->q_end[quadspline_paramPt->ax_dof+k];
    }
    else{
        x0=quadspline_specificPt->q_init[quadspline_paramPt->rz_dof];
        v0=0;
        a0=0;
        xF=quadspline_specificPt->q_end[quadspline_paramPt->rz_dof];
        vF=0;
        aF=0;
    }
    vmax=quadspline_paramPt->vmax;
    amax=quadspline_paramPt->amax;
    jmax=quadspline_paramPt->jmax;
    smax=quadspline_paramPt->smax;

    dv=0.5;

    tD=0;
    if(x0<xF){
        vMax=vmax+dv;
        do{
            vMax=vMax-dv;
            get_xC_xE(robotPt,quadspline_specificPt,k,vMax,&xC,&xE);
        }while(xE<xC);
        if(vMax<vmax){
            vMaxMin=vMax;
            vMaxMax=vMax+dv;
            do{
                vMax=(vMaxMin+vMaxMax)/2;
                get_xC_xE(robotPt,quadspline_specificPt,k,vMax,&xC,&xE);
                if(xE<xC){
                    vMaxMax=vMax;
                }
                else{
                    vMaxMin=vMax;
                }
            }while(fabs(vMaxMax-vMaxMin)>1e-10);
            vMax=vMaxMin;
            get_xC_xE(robotPt,quadspline_specificPt,k,vMax,&xC,&xE);
        }
        if(fabs(xE-xC)>1e-10){
            tD=(xE-xC)/vMax;
        }
    }
    else{
        vMax=-vmax-dv;
        do{
            vMax=vMax+dv;
            get_xC_xE(robotPt,quadspline_specificPt,k,vMax,&xC,&xE);
        }while(xE>xC);
        if(vMax>-vmax){
            vMaxMin=vMax;
            vMaxMax=vMax-dv;
            do{
                vMax=(vMaxMin+vMaxMax)/2;
                get_xC_xE(robotPt,quadspline_specificPt,k,vMax,&xC,&xE);
                if(xE>xC){
                    vMaxMax=vMax;
                }
                else{
                    vMaxMin=vMax;
                }
            }while(fabs(vMaxMax-vMaxMin)>1e-10);
            vMax=vMaxMin;
            get_xC_xE(robotPt,quadspline_specificPt,k,vMax,&xC,&xE);
        }
        if(fabs(xE-xC)>1e-10){
            tD=(xE-xC)/vMax;
        }
    }

    if(fabs(vMax)>vmax){
        vMax=sign(vMax)*vmax;
        get_xC_xE(robotPt,quadspline_specificPt,k,vMax,&xC,&xE);
        if(fabs(xE-xC)>1e-10){
            tD=(xE-xC)/vMax;
        }
    }

    quadspline_specificPt->times[F][k]+=tD;
    quadspline_specificPt->times[D][k]=tD;

    return vMax;

}

void getTimes(p3d_rob *robotPt, p3d_quadspline_data *quadspline_specificPt,
              int k, double tFdes){

    lm_quadspline_str *quadspline_paramPt;
    double vmax,amax,jmax,smax,x0,xF,v0,vF,a0,aF;
    double vDes,vMin,vMax,xC,xE,tD;

    quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in getTimes : quadspline_paramPt is NULL\n");
        return;
    }

    if(k<3){
        x0=quadspline_specificPt->q_init[quadspline_paramPt->x_dof+k];
        v0=quadspline_specificPt->q_init[quadspline_paramPt->vx_dof+k];
        a0=quadspline_specificPt->q_init[quadspline_paramPt->ax_dof+k];
        xF=quadspline_specificPt->q_end[quadspline_paramPt->x_dof+k];
        vF=quadspline_specificPt->q_end[quadspline_paramPt->vx_dof+k];
        aF=quadspline_specificPt->q_end[quadspline_paramPt->ax_dof+k];
    }
    else{
        x0=quadspline_specificPt->q_init[quadspline_paramPt->rz_dof];
        v0=0;
        a0=0;
        xF=quadspline_specificPt->q_end[quadspline_paramPt->rz_dof];
        vF=0;
        aF=0;
    }

    vmax=quadspline_paramPt->vmax;
    amax=quadspline_paramPt->amax;
    jmax=quadspline_paramPt->jmax;
    smax=quadspline_paramPt->smax;

    vDes=getVopt(robotPt,quadspline_specificPt,k);
    if(tFdes>quadspline_specificPt->times[F][k]){
        vMin=0;
        vMax=vDes;
        do{
            vDes=(vMax+vMin)/2;
            get_xC_xE(robotPt,quadspline_specificPt,k,vDes,&xC,&xE);
            if(fabs(xE-xC)>1e-10){
                tD=(xE-xC)/vDes;
                quadspline_specificPt->times[D][k]=tD;
                quadspline_specificPt->times[F][k]+=tD;
            }
            if(tFdes>quadspline_specificPt->times[F][k]){
                vMax=vDes;
            }
            else{
                vMin=vDes;
            }
        }while(fabs(vMax-vMin)>1e-10);
    }
}

void getSplineAt(p3d_rob *robotPt, p3d_quadspline_data *quadspline_specificPt,
                 int kk, double t, double *x, double *v, double *a){

    lm_quadspline_str *quadspline_paramPt;
    double smax,x0,xF,v0,vF,a0,aF;
    double t_cur,x_cur,v_cur,a_cur,j_cur,t_rel;
    double tA1,tA2,sA,tB,tC1,tC2,sC,tD,tE1,tE2,sE,tG,tH1,tH2,sH,tF;
    double t_2,t_3,t_4;
    double t_rel_2,t_rel_3,t_rel_4;
    double tA1_2,tA1_3,tA1_4;
    double tA2_2,tA2_3;
    double tB_2;
    double tC1_2,tC1_3,tC1_4;
    double tC2_2,tC2_3;
    double tE1_2,tE1_3,tE1_4;
    double tE2_2,tE2_3;
    double tG_2;
    double tH1_2,tH1_3,tH1_4;
    double tH2_2,tH2_3;

    int k=kk;

    quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in getTimes : quadspline_paramPt is NULL\n");
        return;
    }

    if(k>=0){
        if(k<3){
            x0=quadspline_specificPt->q_init[quadspline_paramPt->x_dof+k];
            v0=quadspline_specificPt->q_init[quadspline_paramPt->vx_dof+k];
            a0=quadspline_specificPt->q_init[quadspline_paramPt->ax_dof+k];
            xF=quadspline_specificPt->q_end[quadspline_paramPt->x_dof+k];
            vF=quadspline_specificPt->q_end[quadspline_paramPt->vx_dof+k];
            aF=quadspline_specificPt->q_end[quadspline_paramPt->ax_dof+k];
        }
        else{
            x0=quadspline_specificPt->q_init[quadspline_paramPt->rz_dof];
            v0=0;
            a0=0;
            xF=quadspline_specificPt->q_end[quadspline_paramPt->rz_dof];
            vF=0;
            aF=0;
        }
    }
    else{
        double y0,z0,yF,zF,norm2;
        x0=quadspline_specificPt->q_init[quadspline_paramPt->x_dof];
        y0=quadspline_specificPt->q_init[quadspline_paramPt->y_dof];
        z0=quadspline_specificPt->q_init[quadspline_paramPt->z_dof];
        xF=quadspline_specificPt->q_end[quadspline_paramPt->x_dof];
        yF=quadspline_specificPt->q_end[quadspline_paramPt->y_dof];
        zF=quadspline_specificPt->q_end[quadspline_paramPt->z_dof];
        norm2=sqrt(SQR(xF-x0)+SQR(yF-y0)+SQR(zF-z0));
        k=0;
        x0=0;
        v0=0;
        a0=0;
        xF=norm2;
        vF=0;
        aF=0;
    }

    smax=quadspline_paramPt->smax;

    sA=quadspline_specificPt->signes[A][k];
    sC=quadspline_specificPt->signes[C][k];
    sE=quadspline_specificPt->signes[E][k];
    sH=quadspline_specificPt->signes[H][k];

    tA1=quadspline_specificPt->times[A1][k];
    tA2=quadspline_specificPt->times[A2][k];
    tB=quadspline_specificPt->times[B][k];
    tC1=quadspline_specificPt->times[C1][k];
    tC2=quadspline_specificPt->times[C2][k];
    tD=quadspline_specificPt->times[D][k];
    tE1=quadspline_specificPt->times[E1][k];
    tE2=quadspline_specificPt->times[E2][k];
    tG=quadspline_specificPt->times[G][k];
    tH1=quadspline_specificPt->times[H1][k];
    tH2=quadspline_specificPt->times[H2][k];

    tF=quadspline_specificPt->tf;

    t_2=t*t;
    t_3=t_2*t;
    t_4=t_3*t;

    tA1_2=tA1*tA1;
    tA1_3=tA1_2*tA1;
    tA1_4=tA1_3*tA1;

    tA2_2=tA2*tA2;
    tA2_3=tA2_2*tA2;

    tB_2=tB*tB;

    tC1_2=tC1*tC1;
    tC1_3=tC1_2*tC1;
    tC1_4=tC1_3*tC1;

    tC2_2=tC2*tC2;
    tC2_3=tC2_2*tC2;

    tE1_2=tE1*tE1;
    tE1_3=tE1_2*tE1;
    tE1_4=tE1_3*tE1;

    tE2_2=tE2*tE2;
    tE2_3=tE2_2*tE2;

    tG_2=tG*tG;

    tH1_2=tH1*tH1;
    tH1_3=tH1_2*tH1;
    tH1_4=tH1_3*tH1;

    tH2_2=tH2*tH2;
    tH2_3=tH2_2*tH2;

    t_cur=0;
    x_cur=x0;
    v_cur=v0;
    a_cur=a0;
    j_cur=0;

    if(t<=0){
        *a=a0;
        *v=v0;
        *x=x0;
        return;
    }

    if(t<=tA1){
        *a=sA*smax*t_2/2+a0;
        *v=sA*smax*t_3/6+a0*t+v0;
        *x=sA*smax*t_4/24+a0*t_2/2+v0*t+x0;
        return;
    }

    if(tA1>0){
        t_cur=tA1;
        j_cur=sA*smax*tA1;
        a_cur=sA*smax*tA1_2/2+a0;
        v_cur=sA*smax*tA1_3/6+a0*tA1+v0;
        x_cur=sA*smax*tA1_4/24+a0*tA1_2/2+v0*tA1+x0;
    }

    if(t-t_cur<=tA2){
        t_rel=t-t_cur;
        t_rel_2=t_rel*t_rel;
        t_rel_3=t_rel_2*t_rel;
        *a=j_cur*t_rel+a_cur;
        *v=j_cur*t_rel_2/2+a_cur*t_rel+v_cur;
        *x=j_cur*t_rel_3/6+a_cur*t_rel_2/2+v_cur*t_rel+x_cur;
        return;
    }

    if(tA2>0){
        t_cur=t_cur+tA2;
        x_cur=j_cur*tA2_3/6+a_cur*tA2_2/2+v_cur*tA2+x_cur;
        v_cur=j_cur*tA2_2/2+a_cur*tA2+v_cur;
        a_cur=j_cur*tA2+a_cur;
    }

    if(t-t_cur<=tA1){
        t_rel=t-t_cur;
        t_rel_2=t_rel*t_rel;
        t_rel_3=t_rel_2*t_rel;
        t_rel_4=t_rel_3*t_rel;
        *a=-sA*smax*t_rel_2/2+j_cur*t_rel+a_cur;
        *v=-sA*smax*t_rel_3/6+j_cur*t_rel_2/2+a_cur*t_rel+v_cur;
        *x=-sA*smax*t_rel_4/24+j_cur*t_rel_3/6+a_cur*t_rel_2/2+v_cur*t_rel
          +x_cur;
        return;
    }

    if(tA1>0){
        t_cur=t_cur+tA1;
        x_cur=-sA*smax*tA1_4/24+j_cur*tA1_3/6+a_cur*tA1_2/2+v_cur*tA1+x_cur;
        v_cur=-sA*smax*tA1_3/6+j_cur*tA1_2/2+a_cur*tA1+v_cur;
        a_cur=-sA*smax*tA1_2/2+j_cur*tA1+a_cur;
        j_cur=-sA*smax*tA1+j_cur;
    }

    if(t-t_cur<=tB){
        t_rel=t-t_cur;
        t_rel_2=t_rel*t_rel;
        t_rel_3=t_rel_2*t_rel;
        *a=a_cur;
        *v=a_cur*t_rel+v_cur;
        *x=a_cur*t_rel_2/2+v_cur*t_rel+x_cur;
        return;
    }

    if(tB>0){
        t_cur=t_cur+tB;
        x_cur=a_cur*tB_2/2+v_cur*tB+x_cur;
        v_cur=a_cur*tB+v_cur;
    }

    if(t-t_cur<=tC1){
        t_rel=t-t_cur;
        t_rel_2=t_rel*t_rel;
        t_rel_3=t_rel_2*t_rel;
        t_rel_4=t_rel_3*t_rel;
        *a=sC*smax*t_rel_2/2+j_cur*t_rel+a_cur;
        *v=sC*smax*t_rel_3/6+j_cur*t_rel_2/2+a_cur*t_rel+v_cur;
        *x=sC*smax*t_rel_4/24+j_cur*t_rel_3/6+a_cur*t_rel_2/2+v_cur*t_rel+x_cur;
        return;
    }

    if(tC1>0){
        t_cur=t_cur+tC1;
        x_cur=sC*smax*tC1_4/24+j_cur*tC1_3/6+a_cur*tC1_2/2+v_cur*tC1+x_cur;
        v_cur=sC*smax*tC1_3/6+j_cur*tC1_2/2+a_cur*tC1+v_cur;
        a_cur=sC*smax*tC1_2/2+j_cur*tC1+a_cur;
        j_cur=sC*smax*tC1+j_cur;
    }

    if(t-t_cur<=tC2){
        t_rel=t-t_cur;
        t_rel_2=t_rel*t_rel;
        t_rel_3=t_rel_2*t_rel;
        *a=j_cur*t_rel+a_cur;
        *v=j_cur*t_rel_2/2+a_cur*t_rel+v_cur;
        *x=j_cur*t_rel_3/6+a_cur*t_rel_2/2+v_cur*t_rel+x_cur;
        return;
    }

    if(tC2>0){
        t_cur=t_cur+tC2;
        x_cur=j_cur*tC2_3/6+a_cur*tC2_2/2+v_cur*tC2+x_cur;
        v_cur=j_cur*tC2_2/2+a_cur*tC2+v_cur;
        a_cur=j_cur*tC2+a_cur;
    }

    if(t-t_cur<=tC1){
        t_rel=t-t_cur;
        t_rel_2=t_rel*t_rel;
        t_rel_3=t_rel_2*t_rel;
        t_rel_4=t_rel_3*t_rel;
        *a=-sC*smax*t_rel_2/2+j_cur*t_rel+a_cur;
        *v=-sC*smax*t_rel_3/6+j_cur*t_rel_2/2+a_cur*t_rel+v_cur;
        *x=-sC*smax*t_rel_4/24+j_cur*t_rel_3/6+a_cur*t_rel_2/2+v_cur*t_rel+x_cur;
        return;
    }

    if(tC1>0){
        t_cur=t_cur+tC1;
        x_cur=-sC*smax*tC1_4/24+j_cur*tC1_3/6+a_cur*tC1_2/2+v_cur*tC1+x_cur;
        v_cur=-sC*smax*tC1_3/6+j_cur*tC1_2/2+a_cur*tC1+v_cur;
        a_cur=-sC*smax*tC1_2/2+j_cur*tC1+a_cur;
        j_cur=-sC*smax*tC1+j_cur;
    }

    if(t-t_cur<=tD){
        t_rel=t-t_cur;
        t_rel_2=t_rel*t_rel;
        t_rel_3=t_rel_2*t_rel;
        *a=0;
        *v=v_cur;
        *x=v_cur*t_rel+x_cur;
        return;
    }

    t_cur=tF;
    x_cur=xF;
    v_cur=vF;
    a_cur=aF;
    j_cur=0;

    if(t>=t_cur-tH1){
        t_rel=t-(t_cur-tH1);
        t_rel_2=t_rel*t_rel;
        t_rel_3=t_rel_2*t_rel;
        t_rel_4=t_rel_3*t_rel;
        *a=-sH*smax*t_rel_2/2+sH*smax*tH1*t_rel+a_cur-(sH*smax*tH1_2)/2;
        *v=-sH*smax*t_rel_3/6+sH*smax*tH1*t_rel_2/2
          +(a_cur-(sH*smax*tH1_2)/2)*t_rel+(sH*smax*tH1_3)/6-a_cur*tH1+v_cur;
        *x=-sH*smax*t_rel_4/24+sH*smax*tH1*t_rel_3/6
          +(a_cur-(sH*smax*tH1_2)/2)*t_rel_2/2
          +((sH*smax*tH1_3)/6-a_cur*tH1+v_cur)*t_rel-(sH*smax*tH1_4)/24
          +(a_cur*tH1_2)/2-v_cur*tH1+x_cur;
        return;
    }

    if(tH1>0){
        t_cur=t_cur-tH1;
        x_cur=-(sH*smax*tH1_4)/24+(a_cur*tH1_2)/2-v_cur*tH1+x_cur;
        v_cur=(sH*smax*tH1_3)/6-a_cur*tH1+v_cur;
        a_cur=a_cur-(sH*smax*tH1_2)/2;
        j_cur=sH*smax*tH1;
    }

    if(t>=t_cur-tH2){
        t_rel=t-(t_cur-tH2);
        t_rel_2=t_rel*t_rel;
        t_rel_3=t_rel_2*t_rel;
        *a=j_cur*t_rel+a_cur-j_cur*tH2;
        *v=j_cur*t_rel_2/2+(a_cur-j_cur*tH2)*t_rel+(j_cur*tH2_2)/2-a_cur*tH2
          +v_cur;
        *x=j_cur*t_rel_3/6+(a_cur-j_cur*tH2)*t_rel_2/2
          +((j_cur*tH2_2)/2-a_cur*tH2+v_cur)*t_rel-(j_cur*tH2_3)/6
          +(a_cur*tH2_2)/2-v_cur*tH2+x_cur;
        return;
    }

    if(tH2>0){
        t_cur=t_cur-tH2;
        x_cur=-(j_cur*tH2_3)/6+(a_cur*tH2_2)/2-v_cur*tH2+x_cur;
        v_cur=(j_cur*tH2_2)/2-a_cur*tH2+v_cur;
        a_cur=a_cur-j_cur*tH2;
    }

    if(t>=t_cur-tH1){
        t_rel=t-(t_cur-tH1);
        t_rel_2=t_rel*t_rel;
        t_rel_3=t_rel_2*t_rel;
        t_rel_4=t_rel_3*t_rel;
        *a=sH*smax*t_rel_2/2+a_cur-(sH*smax*tH1_2)/2;
        *v=sH*smax*t_rel_3/6+(a_cur-(sH*smax*tH1_2)/2)*t_rel+(sH*smax*tH1_3)/3
          -a_cur*tH1+v_cur;
        *x=sH*smax*t_rel_4/24+(a_cur-(sH*smax*tH1_2)/2)*t_rel_2/2
          +((sH*smax*tH1_3)/3-a_cur*tH1+v_cur)*t_rel-(sH*smax*tH1_4)/8
          +(a_cur*tH1_2)/2-v_cur*tH1+x_cur;
        return;
    }

    if(tH1>0){
        t_cur=t_cur-tH1;
        x_cur=-(sH*smax*tH1_4)/8+(a_cur*tH1_2)/2-v_cur*tH1+x_cur;
        v_cur=(sH*smax*tH1_3)/3-a_cur*tH1+v_cur;
        a_cur=a_cur-(sH*smax*tH1_2)/2;
        j_cur=0;
    }

    if(t>=t_cur-tG){
        t_rel=t-(t_cur-tG);
        t_rel_2=t_rel*t_rel;
        t_rel_3=t_rel_2*t_rel;
        *a=a_cur;
        *v=a_cur*t_rel+v_cur-a_cur*tG;
        *x=a_cur*t_rel_2/2+(v_cur-a_cur*tG)*t_rel+(a_cur*tG_2)/2-v_cur*tG+x_cur;
        return;
    }

    if(tG>0){
        t_cur=t_cur-tG;
        x_cur=(a_cur*tG_2)/2-v_cur*tG+x_cur;
        v_cur=v_cur-a_cur*tG;
    }

    if(t>=t_cur-tE1){
        t_rel=t-(t_cur-tE1);
        t_rel_2=t_rel*t_rel;
        t_rel_3=t_rel_2*t_rel;
        t_rel_4=t_rel_3*t_rel;
        *a=sE*smax*t_rel_2/2-sE*smax*tE1*t_rel+(sE*smax*tE1_2)/2+a_cur;
        *v=sE*smax*t_rel_3/6-sE*smax*tE1*t_rel_2/2
          +((sE*smax*tE1_2)/2+a_cur)*t_rel+v_cur-a_cur*tE1-(sE*smax*tE1_3)/6;
        *x=sE*smax*t_rel_4/24-sE*smax*tE1*t_rel_3/6
          +((sE*smax*tE1_2)/2+a_cur)*t_rel_2/2
          +(v_cur-a_cur*tE1-(sE*smax*tE1_3)/6)*t_rel+(sE*smax*tE1_4)/24
          +(a_cur*tE1_2)/2-v_cur*tE1+x_cur;
        return;
    }

    if(tE1>0){
        t_cur=t_cur-tE1;
        x_cur=(sE*smax*tE1_4)/24+(a_cur*tE1_2)/2-v_cur*tE1+x_cur;
        v_cur=v_cur-a_cur*tE1-(sE*smax*tE1_3)/6;
        a_cur=(sE*smax*tE1_2)/2+a_cur;
        j_cur=-sE*smax*tE1;
    }

    if(t>=t_cur-tE2){
        t_rel=t-(t_cur-tE2);
        t_rel_2=t_rel*t_rel;
        t_rel_3=t_rel_2*t_rel;
        *a=j_cur*t_rel+a_cur-j_cur*tE2;
        *v=j_cur*t_rel_2/2+(a_cur-j_cur*tE2)*t_rel+(j_cur*tE2_2)/2-a_cur*tE2
         +v_cur;
        *x=j_cur*t_rel_3/6+(a_cur-j_cur*tE2)*t_rel_2/2
         +((j_cur*tE2_2)/2-a_cur*tE2+v_cur)*t_rel-(j_cur*tE2_3)/6
         +(a_cur*tE2_2)/2-v_cur*tE2+x_cur;
        return;
    }

    if(tE2>0){
        t_cur=t_cur-tE2;
        x_cur=-(j_cur*tE2_3)/6+(a_cur*tE2_2)/2-v_cur*tE2+x_cur;
        v_cur=(j_cur*tE2_2)/2-a_cur*tE2+v_cur;
        a_cur=a_cur-j_cur*tE2;
    }

    if(t>=t_cur-tE1){
        t_rel=t-(t_cur-tE1);
        t_rel_2=t_rel*t_rel;
        t_rel_3=t_rel_2*t_rel;
        t_rel_4=t_rel_3*t_rel;
        *a=-sE*smax*t_rel_2/2+(sE*smax*tE1_2)/2+a_cur;
        *v=-sE*smax*t_rel_3/6+((sE*smax*tE1_2)/2+a_cur)*t_rel+v_cur-a_cur*tE1
          -(sE*smax*tE1_3)/3;
        *x=-sE*smax*t_rel_4/24+((sE*smax*tE1_2)/2+a_cur)*t_rel_2/2
          +(v_cur-a_cur*tE1-(sE*smax*tE1_3)/3)*t_rel+(sE*smax*tE1_4)/8
          +(a_cur*tE1_2)/2-v_cur*tE1+x_cur;
        return;
    }

}

/* allocation of a data structure specific to the quadspline local method */
p3d_quadspline_data * p3d_alloc_spec_quadspline_localpath(p3d_rob *robotPt,
                                                          configPt q_i,
                                                          configPt q_f){

    p3d_quadspline_data * quadspline_data;
    int k;
    double v,tFdes=0;

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_alloc_spec_quadspline_local_path\n");
    }

    lm_quadspline_str *quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in p3d_alloc_spec_quadspline_localpath : \n");
        printf("quadspline_data is NULL\n");
        return NULL;
    }

    if ((quadspline_data = MY_ALLOC(p3d_quadspline_data,1)) == NULL)
        return NULL;

    quadspline_data->q_init = q_i;
    quadspline_data->q_end = q_f;
    quadspline_data->calculated=FALSE;

    if(fabs(q_i[quadspline_paramPt->vx_dof])<1e-10 &&
       fabs(q_i[quadspline_paramPt->vy_dof])<1e-10 &&
       fabs(q_i[quadspline_paramPt->vz_dof])<1e-10 &&
       fabs(q_i[quadspline_paramPt->ax_dof])<1e-10 &&
       fabs(q_i[quadspline_paramPt->ay_dof])<1e-10 &&
       fabs(q_i[quadspline_paramPt->az_dof])<1e-10 &&
       fabs(q_f[quadspline_paramPt->vx_dof])<1e-10 &&
       fabs(q_f[quadspline_paramPt->vy_dof])<1e-10 &&
       fabs(q_f[quadspline_paramPt->vz_dof])<1e-10 &&
       fabs(q_f[quadspline_paramPt->ax_dof])<1e-10 &&
       fabs(q_f[quadspline_paramPt->ay_dof])<1e-10 &&
       fabs(q_f[quadspline_paramPt->az_dof])<1e-10 ){
        double x0,y0,z0,xf,yf,zf,norm2;
        x0=q_i[quadspline_paramPt->x_dof];
        y0=q_i[quadspline_paramPt->y_dof];
        z0=q_i[quadspline_paramPt->z_dof];
        xf=q_f[quadspline_paramPt->x_dof];
        yf=q_f[quadspline_paramPt->y_dof];
        zf=q_f[quadspline_paramPt->z_dof];
        norm2=sqrt(SQR(xf-x0)+SQR(yf-y0)+SQR(zf-z0));
        q_i[quadspline_paramPt->x_dof]=0;
        q_f[quadspline_paramPt->x_dof]=norm2;
        v=getVopt(robotPt,quadspline_data,0);
        tFdes=quadspline_data->times[F][0];
        v=getVopt(robotPt,quadspline_data,3);
        if(quadspline_data->times[F][3]>tFdes)
            tFdes=quadspline_data->times[F][3];
        getTimes(robotPt,quadspline_data,0,tFdes);
        getTimes(robotPt,quadspline_data,3,tFdes);
        q_i[quadspline_paramPt->x_dof]=x0;
        q_f[quadspline_paramPt->x_dof]=xf;
        quadspline_data->tf=tFdes;
    }
    else{
        for(k=0;k<4;k++){
            v=getVopt(robotPt,quadspline_data,k);
            if(quadspline_data->times[F][k]>tFdes)
                tFdes=quadspline_data->times[F][k];
        }
        for(k=0;k<4;k++){
            getTimes(robotPt,quadspline_data,k,tFdes);
        }
        quadspline_data->tf=tFdes;
    }

    quadspline_data->calculated=TRUE;

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_alloc_spec_quadspline_local_path\n");
    }

    return quadspline_data;
}

/* allocation of local path of type quadspline */
p3d_localpath * p3d_alloc_quadspline_localpath(p3d_rob *robotPt,
                    configPt q_i, configPt q_f,
                    int lp_id,int is_valid){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_alloc_quadspline_localpath\n");
    }

    p3d_localpath * localpathPt = NULL;

    if ((localpathPt = MY_ALLOC(p3d_localpath, 1)) == NULL)
        return NULL;

    /* allocation of the specific part */
    localpathPt->specific.quadspline_data = p3d_alloc_spec_quadspline_localpath(
                                                               robotPt,q_i,q_f);

    if (localpathPt->specific.quadspline_data == NULL){
        /* allocation failed free everything and return NULL*/
        MY_FREE(localpathPt, p3d_localpath, 1);
        return NULL;
    }

    /* Modification of the attitude of q_i and q_f */

    double Tx,Ty,Tz,T,m,f,g,ry,rx,vx,vy,vz,ax,ay,az;

    lm_quadspline_str *quadspline_paramPt;
    quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in p3d_alloc_quadspline_localpath : ");
        printf("quadspline_paramPt is NULL\n");
        return 0;
    }

    m=quadspline_paramPt->mass;
    f=quadspline_paramPt->friction;
    g=quadspline_paramPt->gravity;

    vx=q_i[quadspline_paramPt->vx_dof];
    vy=q_i[quadspline_paramPt->vy_dof];
    vz=q_i[quadspline_paramPt->vz_dof];

    ax=q_i[quadspline_paramPt->ax_dof];
    ay=q_i[quadspline_paramPt->ay_dof];
    az=q_i[quadspline_paramPt->az_dof];

    Tx=m*ax+f*vx;
    Ty=m*ay+f*vy;
    Tz=m*az+f*vz+m*g;
    T=sqrt(SQR(Tx)+SQR(Ty)+SQR(Tz));

    ry=asin(Tx/T);
    rx=asin(-Ty/(T*cos(ry)));

    q_f[quadspline_paramPt->rx_dof]=rx;
    q_f[quadspline_paramPt->ry_dof]=ry;

    vx=q_f[quadspline_paramPt->vx_dof];
    vy=q_f[quadspline_paramPt->vy_dof];
    vz=q_f[quadspline_paramPt->vz_dof];

    ax=q_f[quadspline_paramPt->ax_dof];
    ay=q_f[quadspline_paramPt->ay_dof];
    az=q_f[quadspline_paramPt->az_dof];

    Tx=m*ax+f*vx;
    Ty=m*ay+f*vy;
    Tz=m*az+f*vz+m*g;
    T=sqrt(SQR(Tx)+SQR(Ty)+SQR(Tz));

    ry=asin(Tx/T);
    rx=asin(-Ty/(T*cos(ry)));

    q_f[quadspline_paramPt->rx_dof]=rx;
    q_f[quadspline_paramPt->ry_dof]=ry;

    /* Initialization of the generic part */

    /* fields */
    localpathPt->type_lp = QUADSPLINE;
    localpathPt->valid = is_valid;
    localpathPt->lp_id = lp_id;
    localpathPt->prev_lp = NULL;
    localpathPt->next_lp = NULL;
    localpathPt->length_lp=localpathPt->specific.quadspline_data->tf;

    /* methods associated to the local path */

    /* compute the length of the local path */
    localpathPt->length =
        (double (*)(p3d_rob*, p3d_localpath*))(p3d_quadspline_dist);

    /* computes the configuration at given parameter along the path */
    localpathPt->config_at_param =
        (configPt (*)(p3d_rob*, p3d_localpath*,double))
                                               (p3d_quadspline_config_at_param);
    localpathPt->config_at_distance =
        (configPt (*)(p3d_rob*, p3d_localpath*,double))
                                               (p3d_quadspline_config_at_param);
    localpathPt->extract_sub_localpath =
        (p3d_localpath* (*)(p3d_rob*, p3d_localpath*,double, double))
                                                       (p3d_extract_quadspline);

    /* extract from a local path a sub local path starting from parameter u1
       and ending at parameter u2 */
    localpathPt->extract_by_param =
        (p3d_localpath* (*)(p3d_rob*, p3d_localpath*,double, double))
                                                       (p3d_extract_quadspline);

    /* destroy the localpath */
    localpathPt->destroy =
        (void (*)(p3d_rob*, p3d_localpath*))(p3d_quadspline_destroy);

    /*copy the local path */
    localpathPt->copy =
        (p3d_localpath* (*)(p3d_rob*,p3d_localpath*))
                                                (p3d_copy_quadspline_localpath);

    /* compute the cost of a local path */
    localpathPt->cost =
        (double (*)(p3d_rob*, p3d_localpath*))(p3d_quadspline_cost);

    /* function that simplifies the sequence of two local paths :
       valid only for RS curves */
    localpathPt->simplify =
        (p3d_localpath* (*)(p3d_rob*, p3d_localpath*, int*))
                                                      (p3d_simplify_quadspline);

    /* write the local path in a file */
    localpathPt->write =
        (int (*)(FILE *, p3d_rob*, p3d_localpath*))(p3d_write_quadspline);

    /* from a configuration on a local path, this function computes an
     interval of parameter on the local path on which all the points
     of the robot move by less than the distance given as input.
     The interval is centered on the configuration given as input. The
     function returns the half length of the interval */
    localpathPt->stay_within_dist =
        (double (*)(p3d_rob*, p3d_localpath*,double, whichway, double*))
                                              (p3d_quadspline_stay_within_dist);

    localpathPt->range_param = localpathPt->length_lp;
    localpathPt->ikSol = NULL;

    //save the active constraints
    localpathPt->nbActiveCntrts = 0;
    localpathPt->activeCntrts = NULL;

    #ifdef LIGHT_PLANNER
        localpathPt->isCarryingObject = FALSE;
        for (int i = 0; i < MAX_CARRIED_OBJECTS; i++) {
            localpathPt->carriedObject[i] = NULL;
        }
    #endif

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_alloc_quadspline_localpath\n");
    }

    return localpathPt;
}

/* distance for the quadspline local method
 * The output is actually the time of the motion between qi and qf
 * along the local path
 * */
double p3d_quadspline_dist(p3d_rob *robotPt, p3d_localpath *localpathPt){

    double v,tFdes=0;
    int k;

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_quadspline_dist\n");
    }

    /* cast the pointer to union p3d_lm_specific to a pointer
     to p3d_quadspline_data */
    p3d_quadspline_data *specificPt = localpathPt->specific.quadspline_data;

    /* test whether the type of local path is the expected one */
    if (localpathPt->type_lp != QUADSPLINE){
        printf("p3d_quadspline_dist: Quadspline local path expected\n");
    }

    if(specificPt==NULL && DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : specificPt is NULL in p3d_quadspline_dist\n");
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_quadspline_dist\n");
    }

    if(specificPt->calculated) return specificPt->tf;

    for(k=0;k<4;k++){
        v=getVopt(robotPt,specificPt,k);
        if(specificPt->times[F][k]>tFdes)
            tFdes=specificPt->times[F][k];
    }

    for(k=0;k<4;k++){
        getTimes(robotPt,specificPt,k,tFdes);
    }

    specificPt->tf=tFdes;
    specificPt->calculated=TRUE;

}

/*
 *  Compute the configuration situated at given parameter on the local path.
 *
 *  Input:  the robot, the distance.
 *
 *  Output: the configuration
 */
configPt p3d_quadspline_config_at_param(p3d_rob *robotPt,
                                        p3d_localpath *localpathPt,
                                        double param){

    if(DEBUG_LEVEL!=0){
        printf("ALEX_DEBUG : [START] p3d_quadspline_config_at_param\n");
    }

    p3d_quadspline_data *quadspline_specificPt;
    plm_quadspline_str quadspline_paramPt = lm_get_quadspline_lm_param(robotPt);
    configPt q_init, q_end, q;
    double m,g,f;
    double x,y,z,rx,ry,rz,vx,vy,vz,ax,ay,az;
    double Tx,Ty,Tz,T;
    int k;

    if (localpathPt == NULL)
        return NULL;

    if (localpathPt->type_lp != QUADSPLINE){
        printf("ERROR in p3d_quadspline_config_at_param : ");
        printf("local path must be of type QUADSPLINE\n");
        return NULL;
    }

    quadspline_specificPt = localpathPt->specific.quadspline_data;

    if (quadspline_specificPt == NULL){
        return NULL;
    }

    m=quadspline_paramPt->mass;
    g=quadspline_paramPt->gravity;
    f=quadspline_paramPt->friction;

    q_init = quadspline_specificPt->q_init;
    q_end = quadspline_specificPt->q_end;

    if (param < 0 || localpathPt->length_lp == 0)
        return p3d_copy_config(robotPt,q_init);

    if (param > localpathPt->length_lp)
        return p3d_copy_config(robotPt,q_end);

    q = p3d_alloc_config(robotPt);

    if(fabs(q_init[quadspline_paramPt->vx_dof])<1e-10 &&
       fabs(q_init[quadspline_paramPt->vy_dof])<1e-10 &&
       fabs(q_init[quadspline_paramPt->vz_dof])<1e-10 &&
       fabs(q_init[quadspline_paramPt->ax_dof])<1e-10 &&
       fabs(q_init[quadspline_paramPt->ay_dof])<1e-10 &&
       fabs(q_init[quadspline_paramPt->az_dof])<1e-10 &&
       fabs(q_end[quadspline_paramPt->vx_dof])<1e-10 &&
       fabs(q_end[quadspline_paramPt->vy_dof])<1e-10 &&
       fabs(q_end[quadspline_paramPt->vz_dof])<1e-10 &&
       fabs(q_end[quadspline_paramPt->ax_dof])<1e-10 &&
       fabs(q_end[quadspline_paramPt->ay_dof])<1e-10 &&
       fabs(q_end[quadspline_paramPt->az_dof])<1e-10 ){
        double x0,y0,z0,xf,yf,zf,norm2,dirx,diry,dirz;
        x0=q_init[quadspline_paramPt->x_dof];
        y0=q_init[quadspline_paramPt->y_dof];
        z0=q_init[quadspline_paramPt->z_dof];
        xf=q_end[quadspline_paramPt->x_dof];
        yf=q_end[quadspline_paramPt->y_dof];
        zf=q_end[quadspline_paramPt->z_dof];
        norm2=sqrt(SQR(xf-x0)+SQR(yf-y0)+SQR(zf-z0));
        dirx=(xf-x0)/norm2;
        diry=(yf-y0)/norm2;
        dirz=(zf-z0)/norm2;
        getSplineAt(robotPt,quadspline_specificPt,-1,param,&x,&vx,&ax);
        q[quadspline_paramPt->x_dof]=x0+x*dirx;
        q[quadspline_paramPt->y_dof]=y0+x*diry;
        q[quadspline_paramPt->z_dof]=z0+x*dirz;
        q[quadspline_paramPt->vx_dof]=vx*dirx;
        q[quadspline_paramPt->vy_dof]=vx*diry;
        q[quadspline_paramPt->vz_dof]=vx*dirz;
        q[quadspline_paramPt->ax_dof]=ax*dirx;
        q[quadspline_paramPt->ay_dof]=ax*diry;
        q[quadspline_paramPt->az_dof]=ax*dirz;
    }
    else{
        for(k=0;k<3;k++){
            getSplineAt(robotPt,quadspline_specificPt,k,param,&x,&vx,&ax);
            q[quadspline_paramPt->x_dof+k]=x;
            q[quadspline_paramPt->vx_dof+k]=vx;
            q[quadspline_paramPt->ax_dof+k]=ax;
        }
    }

    x=q[quadspline_paramPt->x_dof];
    y=q[quadspline_paramPt->y_dof];
    z=q[quadspline_paramPt->z_dof];

    vx=q[quadspline_paramPt->vx_dof];
    vy=q[quadspline_paramPt->vy_dof];
    vz=q[quadspline_paramPt->vz_dof];

    ax=q[quadspline_paramPt->ax_dof];
    ay=q[quadspline_paramPt->ay_dof];
    az=q[quadspline_paramPt->az_dof];

    Tx=m*ax+f*vx;
    Ty=m*ay+f*vy;
    Tz=m*az+f*vz+m*g;
    T=sqrt(SQR(Tx)+SQR(Ty)+SQR(Tz));

    ry=asin(Tx/T);
    rx=asin(-Ty/(T*cos(ry)));

    getSplineAt(robotPt,quadspline_specificPt,3,param,&rz,&vx,&ax);

    q[quadspline_paramPt->rx_dof]=rx;
    q[quadspline_paramPt->ry_dof]=ry;
    q[quadspline_paramPt->rz_dof]=rz;

    for(k=0;k<quadspline_paramPt->x_dof;k++){
        q[k]=param*(q_end[k]-q_init[k])/quadspline_specificPt->tf+q_init[k];
    }
    for(k=quadspline_paramPt->az_dof+1;k<robotPt->nb_dof;k++){
        q[k]=param*(q_end[k]-q_init[k])/quadspline_specificPt->tf+q_init[k];
    }

    if(DEBUG_LEVEL!=0){
        printf("ALEX_DEBUG : [END] p3d_quadspline_config_at_param\n");
    }

    return q;
}

/*
 *  Extract from an quadspline local path the sub local path starting
 *  at parameter t1 and ending at parameter t2.
 *  The length of the extracted local path is computed
 *
 *  If t2 > length local path, return end of local path
 */
p3d_localpath *p3d_extract_quadspline(p3d_rob *robotPt,
                                      p3d_localpath *localpathPt,
                                      double t1, double t2){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_extract_quadspline\n");
    }

    configPt q1, q2;
    p3d_localpath *sub_localpathPt;

    q1 = p3d_quadspline_config_at_param(robotPt, localpathPt, t1);
    if (robotPt->cntrt_manager->cntrts != NULL) {
        p3d_set_and_update_this_robot_conf_multisol(robotPt, q1, NULL, 0,
                                                    localpathPt->ikSol);
        p3d_get_robot_config_into(robotPt, &q1);
    }

    q2 = p3d_quadspline_config_at_param(robotPt, localpathPt, t2);
    if (robotPt->cntrt_manager->cntrts != NULL) {
        p3d_set_and_update_this_robot_conf_multisol(robotPt, q2, NULL, 0,
                                                    localpathPt->ikSol);
        p3d_get_robot_config_into(robotPt, &q2);
    }

    sub_localpathPt = p3d_alloc_quadspline_localpath(robotPt, q1, q2, FORWARD,
                                                     localpathPt->valid);

    p3d_copy_iksol(robotPt->cntrt_manager, localpathPt->ikSol,
                   &(sub_localpathPt->ikSol));
    sub_localpathPt->nbActiveCntrts = localpathPt->nbActiveCntrts;
    sub_localpathPt->activeCntrts = MY_ALLOC(int,
                                             sub_localpathPt->nbActiveCntrts);
    for(int i = 0; i < sub_localpathPt->nbActiveCntrts; i++){
        sub_localpathPt->activeCntrts[i] = localpathPt->activeCntrts[i];
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_extract_quadspline\n");
    }

    return sub_localpathPt;
}

/*
 *  lm_get_quadspline_lm_param --
 *
 *  find the first occurrence of quadspline local method parameters.
 */

lm_quadspline_str *lm_get_quadspline_lm_param(p3d_rob* robotPt){

    if(DEBUG_LEVEL>1){
        printf("ALEX_DEBUG : [START] lm_get_quadspline_lm_param\n");
    }

    lm_list_param_str *list_paramPt = robotPt->local_method_params;
    lm_quadspline_str *resultPt = NULL;
    while (list_paramPt!=NULL) {
        if (list_paramPt->lpl_type != QUADSPLINE) {
            list_paramPt = list_paramPt->next;
        }
        else{
            resultPt = (plm_quadspline_str)(list_paramPt->lm_param);
            if(DEBUG_LEVEL>1){
                printf("ALEX_DEBUG : [END] lm_get_quadspline_lm_param\n");
            }
        return resultPt;
        }
    }
    if(DEBUG_LEVEL>1){
        printf("ALEX_DEBUG : [END (NULL)] lm_get_quadspline_lm_param\n");
    }
    return resultPt;
}

/*
 *  lm_create_quadspline
 *
 *  create a Quadspline local method structure
 *
 *  Input: dtab[0,1,2,3,5,6,7] = vmax,amax,jmax,smax,mass,friction,gravity
 *         itab[0,1] = num_joint
 */

lm_quadspline_str *lm_create_quadspline(p3d_rob *robotPt, double *dtab,
                                        int *itab){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] lm_create_quadspline\n");
    }

    lm_quadspline_str* quadspline_paramPt = NULL;

    if ((quadspline_paramPt = MY_ALLOC(lm_quadspline_str, 1)) == NULL) {
        printf("ERROR in lm_create_quadspline : allocation failed\n");
        return (NULL);
    }

    if(dtab[0]<=0){
        printf("ERROR in lm_create_quadspline : vmax is negative or null\n");
        return NULL;
    }
    quadspline_paramPt->vmax = dtab[0];

    if(dtab[1]<=0){
        printf("ERROR in lm_create_quadspline : amax is negative or null\n");
        return NULL;
    }
    quadspline_paramPt->amax = dtab[1];

    if(dtab[2]<=0){
        printf("ERROR in lm_create_quadspline : jmax is negative or null\n");
        return NULL;
    }
    quadspline_paramPt->jmax = dtab[2];

    if(dtab[3]<=0){
        printf("ERROR in lm_create_quadspline : smax is negative or null\n");
        return NULL;
    }
    quadspline_paramPt->smax = dtab[3];

    if(dtab[4]<=0){
        printf("ERROR in lm_create_quadspline : mass is negative or null\n");
        return NULL;
    }
    quadspline_paramPt->mass = dtab[4];

    if(dtab[5]<=0){
        printf("ERROR in lm_create_quadspline : ");
        printf("friction is negative or null\n");
        return NULL;
    }
    quadspline_paramPt->friction = dtab[5];

    if(dtab[6]<=0){
        printf("ERROR in lm_create_quadspline : gravity is negative or null\n");
        return NULL;
    }
    quadspline_paramPt->gravity = dtab[6];

    if(itab[0]<0 || itab[0]>robotPt->njoints){
        printf("ERROR in lm_create_quadspline : num_joint is out of range\n");
        return NULL;
    }
    quadspline_paramPt->num_joint = itab[0];

    p3d_jnt* jnt=robotPt->joints[quadspline_paramPt->num_joint];
    if(jnt->type!=P3D_QUADROTOR){
        printf("ERROR in lm_create_quadspline : ");
        printf("joint of type P3D_QUADROTOR expected\n");
        return NULL;
    }

    quadspline_paramPt->x_dof=jnt->index_dof;
    quadspline_paramPt->y_dof=jnt->index_dof+1;
    quadspline_paramPt->z_dof=jnt->index_dof+2;
    quadspline_paramPt->rx_dof=jnt->index_dof+3;
    quadspline_paramPt->ry_dof=jnt->index_dof+4;
    quadspline_paramPt->rz_dof=jnt->index_dof+5;
    quadspline_paramPt->vx_dof=jnt->index_dof+6;
    quadspline_paramPt->vy_dof=jnt->index_dof+7;
    quadspline_paramPt->vz_dof=jnt->index_dof+8;
    quadspline_paramPt->ax_dof=jnt->index_dof+9;
    quadspline_paramPt->ay_dof=jnt->index_dof+10;
    quadspline_paramPt->az_dof=jnt->index_dof+11;


    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] lm_create_quadspline\n");
    }

    return(quadspline_paramPt);
}

/*
 *  p3d_create_quadspline_local_method --
 *
 *  create data-structure to store parameters of quadspline local method
 */
int p3d_create_quadspline_local_method(double *dtab, int *itab){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_create_quadspline_local_method\n");
    }

    p3d_rob *robotPt = (p3d_rob*)p3d_get_desc_curid(P3D_ROBOT);

    plm_quadspline_str quadspline_paramPt = lm_get_quadspline_lm_param(robotPt);

    /* test that quadspline local method has not been already initialized
    for this robot */
    if (quadspline_paramPt != NULL) {
        printf(("WARNING : quadspline already initialized\n"));
        return FALSE;
    }

    robotPt->lpl_type = QUADSPLINE;

    quadspline_paramPt = lm_create_quadspline(robotPt, dtab, itab);

    if (quadspline_paramPt != NULL) {
        robotPt->local_method_params =
            lm_append_to_list(robotPt->local_method_params,
                              (void*)quadspline_paramPt, QUADSPLINE);
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_create_quadspline_local_method\n");
    }

    return TRUE;
}

/*
 * destroys a structure of type p3d_quadspline_data
 */
void p3d_destroy_quadspline_data(p3d_rob* robotPt,
                                 p3d_quadspline_data* quadspline_dataPt){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_destroy_quadspline_data\n");
    }

    if (quadspline_dataPt != NULL){
        if (quadspline_dataPt->q_init != NULL){
            p3d_destroy_config(robotPt, quadspline_dataPt->q_init);
        }
        if (quadspline_dataPt->q_end != NULL){
            p3d_destroy_config(robotPt, quadspline_dataPt->q_end);
        }
        MY_FREE(quadspline_dataPt, p3d_quadspline_data, 1);
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_destroy_quadspline_data\n");
    }

}

/*
 * destroy a Quadspline local path
 */
void p3d_quadspline_destroy(p3d_rob* robotPt, p3d_localpath* localpathPt){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_quadspline_destroy\n");
    }

    if (localpathPt != NULL){
        /* test whether the type of local path is the expected one */
        if (localpathPt->type_lp != QUADSPLINE){
            printf("ERROR in p3d_quadspline_destroy : ");
            printf("Quadspline local path expected\n");
        }
        /* destroy the specific part */
        if (localpathPt->specific.quadspline_data != NULL){
            p3d_destroy_quadspline_data(robotPt,
                                        localpathPt->specific.quadspline_data);
        }
        localpathPt->next_lp = NULL;
        localpathPt->prev_lp = NULL;
        if (localpathPt->ikSol){
            p3d_destroy_specific_iksol(robotPt->cntrt_manager,
                                       localpathPt->ikSol);
            localpathPt->ikSol = NULL;
        }
        MY_FREE(localpathPt->activeCntrts, int, localpathPt->nbActiveCntrts);
        MY_FREE(localpathPt, p3d_localpath, 1);
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_quadspline_destroy\n");
    }

}

/*
 *  Copy one local path.
 *
 *  Input:  the robot, the local path.
 *
 *  Output: the copied local path
 */

p3d_localpath *p3d_copy_quadspline_localpath(p3d_rob* robotPt,
                                             p3d_localpath* localpathPt){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_copy_quadspline_localpath\n");
    }

    configPt q_i, q_f;
    q_i=p3d_copy_config(robotPt, localpathPt->specific.quadspline_data->q_init);
    q_f=p3d_copy_config(robotPt, localpathPt->specific.quadspline_data->q_end);

    p3d_localpath* copy_localpathPt=p3d_alloc_quadspline_localpath(robotPt,q_i,
                                                                   q_f,0,TRUE);

    /* update length and range of parameter */
    copy_localpathPt->length_lp = localpathPt->specific.quadspline_data->tf;
    copy_localpathPt->range_param = localpathPt->specific.quadspline_data->tf;
    p3d_copy_iksol(robotPt->cntrt_manager, localpathPt->ikSol,
                   &(copy_localpathPt->ikSol));
    copy_localpathPt->nbActiveCntrts = localpathPt->nbActiveCntrts;
    copy_localpathPt->activeCntrts = MY_ALLOC(int,
                                              copy_localpathPt->nbActiveCntrts);
    for(int i = 0; i < copy_localpathPt->nbActiveCntrts; i++){
        copy_localpathPt->activeCntrts[i] = localpathPt->activeCntrts[i];
    }

    #ifdef LIGHT_PLANNER
        copy_localpathPt->isCarryingObject = localpathPt->isCarryingObject;
        for (int i = 0; i < MAX_CARRIED_OBJECTS; i++) {
            copy_localpathPt->carriedObject[i] = localpathPt->carriedObject[i];
        }
    #endif

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_copy_quadspline_localpath\n");
    }

    return copy_localpathPt;
}

/*
 *  Cost of a local path
 *
 *  Input:  the local path
 *
 *  Output: the cost
 */
double p3d_quadspline_cost(p3d_rob *robotPt, p3d_localpath *localpathPt){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_quadspline_cost\n");
        printf("ALEX_DEBUG : [END] p3d_quadspline_cost\n");
    }

    return localpathPt->length_lp;
}

/*
 *  does nothing
 */
p3d_localpath *p3d_simplify_quadspline(p3d_rob *robotPt,
                               p3d_localpath *localpathPt, int *need_colcheck){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_simplify_quadspline\n");
        printf("ALEX_DEBUG : [END] p3d_simplify_quadspline\n");
    }

    return localpathPt;
}

/*
 *  p3d_write_quadspline --
 *
 *  write a localpath of type quadspline in a file
 *
 *  ARGS IN  : a file descriptor,
 *             a robot,
 *             a localpath
 *
 *  ARGS OUT : TRUE if success,
 *             FALSE if input local path is not an quadspline one.
 */
int p3d_write_quadspline(FILE *file, p3d_rob* robotPt,
                         p3d_localpath* localpathPt){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_write_quadspline\n");
    }

    p3d_quadspline_data *quadspline_dataPt = NULL;

    if (localpathPt->type_lp != QUADSPLINE) {
        return FALSE;
    }

    fprintf(file, "\n\np3d_add_localpath quadspline\n");

    quadspline_dataPt =
                    (pp3d_quadspline_data)localpathPt->specific.quadspline_data;
    /* write each RS segment */
    fprintf(file, "conf_init");
    fprint_config_one_line(file, robotPt, quadspline_dataPt->q_init);
    fprintf(file, "\nconf_end ");
    fprint_config_one_line(file, robotPt, quadspline_dataPt->q_end);
    fprintf(file, "\ntf %f",quadspline_dataPt->tf);
    fprintf(file, "\np3d_end_local_path\n");

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_write_quadspline\n");
    }

    return TRUE;
}

/*  p3d_quadspline_stay_within_dist
 *
 *  Input:  the robot,
 *          the local path,
 *          the parameter along the curve,
 *          the direction of motion
 *          the maximal distance moved by all the points of the
 *          robot
 *
 *  Output: length of the interval of parameter the robot can
 *          stay on without any body moving by more than the input distance
 *
 *  Description:
 *          From a configuration on a local path, this function
 *          computes an interval of parameter on the local path on
 *          which all the points of the robot move by less than the
 *          distance given as input.  The interval is centered on the
 *          configuration given as input. The function returns the
 *          half length of the interval
 */
double p3d_quadspline_stay_within_dist(p3d_rob* robotPt,
                p3d_localpath* localpathPt,
                double parameter, whichway dir,
                double *distances){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_quadspline_stay_within_dist\n");
    }

  p3d_quadspline_data *quadspline_localpathPt=NULL;
  int i, j, njnt = robotPt->njoints;
  p3d_jnt *cur_jntPt, *prev_jntPt;
  configPt q_max_param, q_param;
  double max_param, min_param;
  double range_param = localpathPt->length_lp;
  p3d_stay_within_dist_data * stay_within_dist_data;

  /* local path has to be of type quadspline */
  if (localpathPt->type_lp != QUADSPLINE){
    printf("ERROR in p3d_quadspline_stay_within_dist : ");
    printf("Quadspline local path expected\n");
    return 0;
  }
  //on recupere les donnees du localpath
  quadspline_localpathPt = localpathPt->specific.quadspline_data;

  /* store the data to compute the maximal velocities at the
     joint for each body of the robot */
  stay_within_dist_data = MY_ALLOC(p3d_stay_within_dist_data, njnt+2);
  p3d_init_stay_within_dist_data(stay_within_dist_data);

  if (dir == FORWARD){
    q_max_param = quadspline_localpathPt->q_end;
    min_param = max_param = range_param - parameter;
  }
  else{
    q_max_param = quadspline_localpathPt->q_init;
    min_param = max_param = parameter;
  }
  /* Get the current config to have the modifications of the constraints */
  /* Supose that q_init and q_goal respect constraints */
  q_param = p3d_get_robot_config(robotPt);
  //q_param = localpathPt->config_at_param(robotPt,localpathPt,parameter);

  /* computation of the bounds for the linear and angular
     velocities of each body */

  int minJnt = 0;
  for(i=0; i<=njnt; i++) {
    cur_jntPt = robotPt->joints[i];
    prev_jntPt = cur_jntPt->prev_jnt;

    /* j = index of the joint to which the current joint is attached */
    if (prev_jntPt==NULL)
      { j = -1; } /* environment */
    else
      { j = prev_jntPt->num; }
    double bakMinParam = min_param;

    p3d_jnt_stay_within_dist(&(stay_within_dist_data[j+1]), cur_jntPt,
                 &(stay_within_dist_data[i+1]), &(distances[i]),
                 q_param, q_max_param, max_param, &min_param);
    if (min_param < bakMinParam){
      minJnt = cur_jntPt->num;
    }
  /* Rem: stay_within_dist_data[0] is bound to the environment */
  }

  MY_FREE(stay_within_dist_data, p3d_stay_within_dist_data, njnt+2);
  p3d_destroy_config(robotPt, q_param);

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_quadspline_stay_within_dist\n");
    }

  return min_param;
}

p3d_localpath *p3d_quadspline_localplanner( p3d_rob *robotPt, configPt qi, configPt qf, int * ikSol){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_quadspline_local_planner(with speed)\n");
    }

    configPt initconfPt, goalconfPt;
    p3d_localpath *localpathPt;

    /* on verifie que les configurations de depart et d'arrivee existent */
    if(qi == NULL){
        PrintInfo((("MP: p3d_quadspline_localplanner: no start configuration...\n")));
        p3d_set_search_status(P3D_ILLEGAL_START);
        return(NULL);
    }
    if(qf == NULL){
        PrintInfo((("MP: p3d_quadspline_localplanner: no goal configuration...\n")));
        p3d_set_search_status(P3D_ILLEGAL_GOAL);
        return(NULL);
    }

    /* copy of qi into initconfPt and of qf into goalconfPt */
    initconfPt = p3d_copy_config(robotPt, qi);
    goalconfPt = p3d_copy_config(robotPt, qf);
    if(p3d_get_search_verbose()){
      PrintInfo(("MP: p3d_quadspline_localplanner : "));
      PrintInfo(("qi=("));
      //print_config(robotPt, initconfPt);
      PrintInfo((") ; "));
      PrintInfo(("qf=("));
      //print_config(robotPt, goalconfPt);
      PrintInfo((")\n"));
    }

    /* If initconfPt == goalconfPt, free initconfPt and goalconfPt
     and return NULL */
    if(p3d_equal_config(robotPt,initconfPt, goalconfPt)) {
        PrintInfo((("MP: p3d_quadspline_localplanner: q_init = q_goal!\n")));
        p3d_set_search_status(P3D_CONFIG_EQUAL);
        p3d_destroy_config(robotPt, initconfPt);
        p3d_destroy_config(robotPt, goalconfPt);
        return(NULL);
    }

    localpathPt = p3d_alloc_quadspline_localpath(robotPt,initconfPt,goalconfPt,0,TRUE);

    p3d_set_search_status(P3D_SUCCESS);

    localpathPt->ikSol = ikSol;
    localpathPt->activeCntrts = p3d_getActiveCntrts(robotPt,&(localpathPt->nbActiveCntrts));

    #if defined(LIGHT_PLANNER)
      localpathPt->isCarryingObject = robotPt->isCarryingObject;
      for (int i = 0; i < MAX_CARRIED_OBJECTS; i++){
        if(i < robotPt->armManipulationData->size() ){
          localpathPt->carriedObject[i] = (*robotPt->armManipulationData)[i].getCarriedObject(); /*!< pointer to the carried object (obstacle environment or robot body) */
        }
      }
        //p3d_mat4Copy(robotPt->Tgrasp, localpathMg->Tgrasp);
    #endif

        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] p3d_quadspline_local_planner(with speed)\n");
        }

      //print_quadspline(robotPt,localpathPt);
      return(localpathPt);
}

void print_quadspline(p3d_rob *robotPt,p3d_localpath *localpathPt){
    lm_quadspline_str *quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in print_quadspline : quadspline_data is NULL\n");
    }
    p3d_quadspline_data *quadspline_data=NULL;
    if(localpathPt) quadspline_data=localpathPt->specific.quadspline_data;
    if(quadspline_data==NULL){
        printf("ERROR in print_quadspline : quadspline_data is NULL\n");
    }
    if(quadspline_paramPt){
    printf("Quadspline ----------------------------------------------------\n");
    printf("   vmax=%f\n",quadspline_paramPt->vmax);
    printf("   amax=%f\n",quadspline_paramPt->amax);
    printf("   jmax=%f\n",quadspline_paramPt->jmax);
    printf("   smax=%f\n",quadspline_paramPt->smax);
    printf("   num_joint=%d\n",quadspline_paramPt->num_joint);
    printf("   x_dof=%d\n",quadspline_paramPt->x_dof);
    printf("   y_dof=%d\n",quadspline_paramPt->y_dof);
    printf("   z_dof=%d\n",quadspline_paramPt->z_dof);
    printf("   rx_dof=%d\n",quadspline_paramPt->rx_dof);
    printf("   ry_dof=%d\n",quadspline_paramPt->ry_dof);
    printf("   rz_dof=%d\n",quadspline_paramPt->rz_dof);
    printf("   vx_dof=%d\n",quadspline_paramPt->vx_dof);
    printf("   vy_dof=%d\n",quadspline_paramPt->vy_dof);
    printf("   vz_dof=%d\n",quadspline_paramPt->vz_dof);
    printf("   ax_dof=%d\n",quadspline_paramPt->ax_dof);
    printf("   ay_dof=%d\n",quadspline_paramPt->ay_dof);
    printf("   az_dof=%d\n",quadspline_paramPt->az_dof);
    printf("   m=%f\n",quadspline_paramPt->mass);
    printf("   f=%f\n",quadspline_paramPt->friction);
    printf("   g=%f\n",quadspline_paramPt->gravity);
    if(quadspline_data){
    printf("   calculated=%d\n",quadspline_data->calculated);
    printf("   q_init=[%f ",quadspline_data->q_init[quadspline_paramPt->x_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->y_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->z_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->rx_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->ry_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->rz_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->vx_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->vy_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->vz_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->ax_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->ay_dof]);
    printf("%f]\n",quadspline_data->q_init[quadspline_paramPt->az_dof]);
    printf("   q_end=[%f ",quadspline_data->q_end[quadspline_paramPt->x_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->y_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->z_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->rx_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->ry_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->rz_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->vx_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->vy_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->vz_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->ax_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->ay_dof]);
    printf("%f]\n",quadspline_data->q_end[quadspline_paramPt->az_dof]);
    if(quadspline_data->calculated){
        printf("   signes=[ ");
        for(int k=0;k<4;k++){
            for(int i=0;i<4;i++){
                printf("%d ",(int)quadspline_data->signes[i][k]);
            }
            printf("]\n            ");
        }
        printf("\n   times=[ ");
        for(int k=0;k<4;k++){
            for(int i=0;i<11;i++){
                printf("%f ",quadspline_data->times[i][k]);
            }
            printf("\n           ");
        }
        printf("\n   tf=%f\n",quadspline_data->tf);
    }
    }
    printf("---------------------------------------------------------------\n");
    }
}

void set_qf_voro_opt(p3d_rob *robotPt,p3d_localpath *localpathPt,double deltaPath,double delta){
    lm_quadspline_str *quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in set_qf_voro_opt : quadspline_data is NULL\n");
        return;
    }
    p3d_quadspline_data *quadspline_data=localpathPt->specific.quadspline_data;
    if(quadspline_data==NULL){
        printf("ERROR in set_qf_voro_opt : quadspline_data is NULL\n");
        return;
    }
    int x_dof=quadspline_paramPt->x_dof;
    int y_dof=quadspline_paramPt->y_dof;
    int z_dof=quadspline_paramPt->z_dof;
    int vx_dof=quadspline_paramPt->vx_dof;
    int vy_dof=quadspline_paramPt->vy_dof;
    int vz_dof=quadspline_paramPt->vz_dof;
    int ax_dof=quadspline_paramPt->ax_dof;
    int ay_dof=quadspline_paramPt->ay_dof;
    int az_dof=quadspline_paramPt->az_dof;
    double v_max=quadspline_paramPt->vmax;
    double a_max=quadspline_paramPt->amax;
    configPt q_i=quadspline_data->q_init;
    configPt q_f=quadspline_data->q_end;
    double norm_QfQi=sqrt(SQR(q_f[x_dof]-q_i[x_dof])+SQR(q_f[y_dof]-q_i[y_dof])+SQR(q_f[z_dof]-q_i[z_dof]));
    double norm_A=sqrt(SQR(q_f[ax_dof])+SQR(q_f[ay_dof])+SQR(q_f[az_dof]));
    p3d_point dirV;
    dirV.x=(q_f[x_dof]-q_i[x_dof])/norm_QfQi;
    dirV.y=(q_f[y_dof]-q_i[y_dof])/norm_QfQi;
    dirV.z=(q_f[z_dof]-q_i[z_dof])/norm_QfQi;
    q_f[x_dof]=q_i[x_dof]+deltaPath*dirV.x;
    q_f[y_dof]=q_i[y_dof]+deltaPath*dirV.y;
    q_f[z_dof]=q_i[z_dof]+deltaPath*dirV.z;
    p3d_point dirA;
    dirA.x=q_f[ax_dof]/norm_A;
    dirA.y=q_f[ay_dof]/norm_A;
    dirA.z=q_f[az_dof]/norm_A;
    double norm_speed_opt=deltaPath;
    double norm_acceleration_opt=deltaPath;
    if(delta>0){
        double tf_opt_min=P3D_HUGE;
        for(double norm_speed=0;norm_speed<=v_max;norm_speed+=v_max/delta){
            q_f[vx_dof]=norm_speed*dirV.x;
            q_f[vy_dof]=norm_speed*dirV.y;
            q_f[vz_dof]=norm_speed*dirV.z;
            for(double norm_acceleration=0;norm_acceleration<=a_max;norm_acceleration+=a_max/delta){
                q_f[ax_dof]=norm_acceleration*dirA.x;
                q_f[ay_dof]=norm_acceleration*dirA.y;
                q_f[az_dof]=norm_acceleration*dirA.z;
                double tf_opt=0;
                for(int k=0;k<3;k++){
                    getVopt(robotPt,quadspline_data,k);
                    if(quadspline_data->times[F][k]>tf_opt)
                        tf_opt=quadspline_data->times[F][k];
                }
                if(tf_opt<tf_opt_min){
                    tf_opt_min=tf_opt;
                    norm_speed_opt=norm_speed;
                    norm_acceleration_opt=norm_acceleration;
                }
            }
        }
    }
    double tFdes=0;
    quadspline_data->q_init = q_i;
    quadspline_data->q_end = q_f;
    quadspline_data->calculated=FALSE;
    for(int k=0;k<3;k++){
        getVopt(robotPt,quadspline_data,k);
        if(quadspline_data->times[F][k]>tFdes)
            tFdes=quadspline_data->times[F][k];
    }

    for(int k=0;k<3;k++){
        getTimes(robotPt,quadspline_data,k,tFdes);
    }
    quadspline_data->tf=tFdes;
    quadspline_data->calculated=TRUE;
    return;
}

void p3d_switch_to_quad(p3d_rob* robotPt,p3d_traj* traj){
    if(robotPt==NULL){
        printf("ERROR in p3d_switch_to_quad : robotPt is NULL\n");
        return;
    }
    if(traj==NULL){
        printf("ERROR in p3d_switch_to_quad : traj is NULL\n");
        return;
    }
    lm_quadspline_str *quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in p3d_switch_to_quad : quadspline_data is NULL\n");
        return;
    }
    p3d_localpath *localpathPt=traj->courbePt;
    if(localpathPt->type_lp!=LINEAR){
        printf("ERROR in p3d_switch_to_quad : this is not a linear path\n");
        return;
    }
    localpathPt=traj->courbePt;
    p3d_localpath *prevLP=NULL;
    while(localpathPt!=NULL){
        configPt qi=localpathPt->specific.lin_data->q_init;
        configPt qf=localpathPt->specific.lin_data->q_end;
        // Setting velocity and acceleration to zero
        qi[quadspline_paramPt->vx_dof]=0;
        qi[quadspline_paramPt->vy_dof]=0;
        qi[quadspline_paramPt->vz_dof]=0;
        qi[quadspline_paramPt->ax_dof]=0;
        qi[quadspline_paramPt->ay_dof]=0;
        qi[quadspline_paramPt->az_dof]=0;
        qf[quadspline_paramPt->vx_dof]=0;
        qf[quadspline_paramPt->vy_dof]=0;
        qf[quadspline_paramPt->vz_dof]=0;
        qf[quadspline_paramPt->ax_dof]=0;
        qf[quadspline_paramPt->ay_dof]=0;
        qf[quadspline_paramPt->az_dof]=0;
        p3d_localpath *newLP=p3d_quadspline_localplanner(robotPt,qi,qf,robotPt->ikSol);
        if(localpathPt->prev_lp==NULL) traj->courbePt=newLP;
        newLP->prev_lp=prevLP;
        prevLP=newLP;
        localpathPt=localpathPt->next_lp;
    }
    localpathPt=prevLP;
    prevLP=NULL;
    while(localpathPt!=NULL){
        localpathPt->next_lp=prevLP;
        prevLP=localpathPt;
        localpathPt=localpathPt->prev_lp;
    }
}
